# samplers.py
# https://github.com/ddh0/easy-llama/

"""Submodule containing SamplerSettings class and some preset samplers"""

from sys    import maxsize
from typing import Dict


MAX_TEMP = float(maxsize)

class SamplerSettings:
    """
    A SamplerSettings object specifies the the sampling parameters that will be
    used to control text generation
    """

    ParamTypes: Dict[str, type] = {
        'max_len_tokens':    int,
        'temp':              float,
        'top_p':             float,
        'min_p':             float,
        'frequency_penalty': float,
        'presence_penalty':  float,
        'repeat_penalty':    float,
        'top_k':             int
    }

    def __init__(
            self,
            max_len_tokens:    int   = -1,
            temp:              float = 0.8,
            top_p:             float = 0.95,
            min_p:             float = 0.05,
            frequency_penalty: float = 0.0,
            presence_penalty:  float = 0.0,
            repeat_penalty:    float = 1.0,
            top_k:             int   = 40
        ):
        """
        Construct a new SamplerSettings instance
        """

        self.max_len_tokens    = max_len_tokens
        self.temp              = temp
        self.top_p             = top_p
        self.min_p             = min_p
        self.frequency_penalty = frequency_penalty
        self.presence_penalty  = presence_penalty
        self.repeat_penalty    = repeat_penalty
        self.top_k             = top_k

        for sampler_param in SamplerSettings.ParamTypes:
            expected_type = SamplerSettings.ParamTypes[sampler_param]
            actual_type = type(getattr(self, sampler_param))
            if actual_type != expected_type:
                raise TypeError(
                    f"wrong type for SamplerSettings parameter '{sampler_param}'" + \
                    f" - expected {expected_type}, got {actual_type}"
                )
    
    def __repr__(self) -> str:
        repr_str = 'SamplerSettings('
        repr_str += f'max_len_tokens={self.max_len_tokens}, '
        repr_str += f'temp={self.temp}, '
        repr_str += f'top_p={self.top_p}, '
        repr_str += f'min_p={self.min_p}, '
        repr_str += f'frequency_penalty={self.frequency_penalty}, '
        repr_str += f'presence_penalty={self.presence_penalty}, '
        repr_str += f'repeat_penalty={self.repeat_penalty}, '
        repr_str += f'top_k={self.top_k})'
        return repr_str

# most likely token is always chosen
GreedyDecoding = SamplerSettings(
    temp = 0.0,
)

# reflects llama.cpp
DefaultSampling = SamplerSettings()

# reflects llama.cpp before repeat_penalty changes
OldDefaultSampling = SamplerSettings(
    repeat_penalty = 1.1
)

# original distribution
SimpleSampling = SamplerSettings(
    temp = 1.0,
    top_p = 1.0,
    min_p = 0.0,
    top_k = -1
)

# for models with large vocabulary, which tend to run hot
TikTokenSampling = SamplerSettings(
    temp=0.6,
    repeat_penalty=1.1
)

# use min_p as the only filter (weak)
LowMinPSampling = SamplerSettings(
    temp = MAX_TEMP,
    top_p = 1.0,
    min_p = 0.1,
    top_k = -1
)

# use min_p as the only filter (moderate)
MinPSampling = SamplerSettings(
    temp = MAX_TEMP,
    top_p = 1.0,
    min_p = 0.2,
    top_k = -1
)

# use min_p as the only filter (strong)
StrictMinPSampling = SamplerSettings(
    temp = MAX_TEMP,
    top_p = 1.0,
    min_p = 0.5,
    top_k = -1
)

# https://arxiv.org/abs/2210.14140
ContrastiveSearch = SamplerSettings(
    temp = 0.0,
    presence_penalty = 0.4
)

# https://arxiv.org/abs/2210.14140
WarmContrastiveSearch = SamplerSettings(
    temp = 0.0,
    presence_penalty = 0.8
)

# outputs completely random tokens from vocab (useless)
RandomSampling = SamplerSettings(
    temp = MAX_TEMP,
    top_p = 1.0,
    min_p = 0.0,
    top_k = -1
)

# default sampling with reduced temperature
LowTempSampling = SamplerSettings(
    temp = 0.4
)

# default sampling with increased temperature
HighTempSampling = SamplerSettings(
    temp = 1.2
)



# formats.py
# https://github.com/ddh0/easy-llama/

"""Submodule containing various prompt formats used by models"""

from typing import Dict, Union


def wrap(
        prompt: str,
        format: Dict[str, Union[str, list]]
    ) -> str:
    """Wrap a given string in any prompt format for single-turn completion"""
    return (
        format['system_prefix'] +
        format['system_content'] +
        format['system_postfix'] +
        format['user_prefix'] +
        prompt +
        format['user_postfix'] +
        format['bot_prefix']
    )

blank: Dict[str, Union[str, list]] = {
    "system_prefix": "",
    "system_content": "",
    "system_postfix": "",
    "user_prefix": "",
    "user_content": "",
    "user_postfix": "",
    "bot_prefix": "",
    "bot_content": "",
    "bot_postfix": "",
    "stops": [],
}

# https://github.com/tatsu-lab/stanford_alpaca
alpaca: Dict[str, Union[str, list]] = {
    "system_prefix": "",
    "system_content": "Below is an instruction that describes a task. " + \
    "Write a response that appropriately completes the request.",
    "system_postfix": "\n\n",
    "user_prefix": "### Instruction:\n",
    "user_content": "",
    "user_postfix": "\n\n",
    "bot_prefix": "### Response:\n",
    "bot_content": "",
    "bot_postfix": "\n\n",
    "stops": ['###', 'Instruction:', '\n\n\n'],
}

# https://docs.mistral.ai/models/
# As a reference, here is the format used to tokenize instructions during fine-tuning:
# ```
# [START_SYMBOL_ID] + 
# tok("[INST]") + tok(USER_MESSAGE_1) + tok("[/INST]") +
# tok(BOT_MESSAGE_1) + [END_SYMBOL_ID] +
# …
# tok("[INST]") + tok(USER_MESSAGE_N) + tok("[/INST]") +
# tok(BOT_MESSAGE_N) + [END_SYMBOL_ID]
# ```
# In the pseudo-code above, note that the tokenize method should not add a BOS or EOS token automatically, but should add a prefix space.

mistral_instruct: Dict[str, Union[str, list]] = {
    "system_prefix": "",
    "system_content": "",
    "system_postfix": "",
    "user_prefix": " [INST] ",
    "user_content": "",
    "user_postfix": " [/INST]",
    "bot_prefix": "",
    "bot_content": "",
    "bot_postfix": "",
    "stops": [],
}

# https://docs.mistral.ai/platform/guardrailing/
mistral_instruct_safe: Dict[str, Union[str, list]] = {
    "system_prefix": "",
    "system_content": "",
    "system_postfix": "",
    "user_prefix": " [INST] Always assist with care, respect, and truth. " + \
    "Respond with utmost utility yet securely. Avoid harmful, unethical, " + \
    "prejudiced, or negative content. Ensure replies promote fairness and " + \
    "positivity. ",
    "user_content": "",
    "user_postfix": " [/INST]",
    "bot_prefix": "",
    "bot_content": "",
    "bot_postfix": "",
    "stops": [],
}

# https://github.com/openai/openai-python/blob/main/chatml.md
chatml: Dict[str, Union[str, list]] = {
    "system_prefix": "<|im_start|>system\n",
    "system_content": "",
    "system_postfix": "<|im_end|>\n",
    "user_prefix": "<|im_start|>user\n",
    "user_content": "",
    "user_postfix": "<|im_end|>\n",
    "bot_prefix": "<|im_start|>assistant\n",
    "bot_content": "",
    "bot_postfix": "<|im_end|>\n",
    "stops": ['<|im_start|>', '</s>'],
}

# https://huggingface.co/blog/llama2
# system message relaxed to avoid undue refusals
llama2chat: Dict[str, Union[str, list]] = {
    "system_prefix": "[INST] <<SYS>>\n",
    "system_content": "You are a helpful AI assistant.",
    "system_postfix": "\n<</SYS>>\n\n",
    "user_prefix": "",
    "user_content": "",
    "user_postfix": " [/INST]",
    "bot_prefix": " ",
    "bot_content": "",
    "bot_postfix": " [INST] ",
    "stops": ['[INST]', '[/INST]'],
}

# https://github.com/ggerganov/llama.cpp/issues/6747#issuecomment-2065013606
# TODO: better reference
llama3: Dict[str, Union[str, list]] = {
    "system_prefix": "<|start_header_id|>system<|end_header_id|>\n\n",
    "system_content": 'You are a helpful AI assistant called "Llama 3".',
    "system_postfix": "<|eot_id|>\n",
    "user_prefix": "<|start_header_id|>user<|end_header_id|>\n\n",
    "user_content": "",
    "user_postfix": "<|eot_id|>\n",
    "bot_prefix": "<|start_header_id|>assistant<|end_header_id|>\n\n",
    "bot_content": "",
    "bot_postfix": "<|eot_id|>\n",
    "stops": [128001, 128009],
}

# https://github.com/tatsu-lab/stanford_alpaca
alpaca: Dict[str, Union[str, list]] = {
    "system_prefix": "",
    "system_content": "Below is an instruction that describes a task. " + \
    "Write a response that appropriately completes the request.",
    "system_postfix": "\n\n",
    "user_prefix": "### Instruction:\n",
    "user_content": "",
    "user_postfix": "\n\n",
    "bot_prefix": "### Response:\n",
    "bot_content": "",
    "bot_postfix": "\n\n",
    "stops": ['###', 'Instruction:', '\n\n\n'],
}

# this is the official vicuna. it is often butchered in various ways,
# most commonly by adding line breaks
# https://github.com/flu0r1ne/FastChat/blob/main/docs/vicuna_weights_version.md
vicuna_lmsys: Dict[str, Union[str, list]] = {
    "system_prefix": "",
    "system_content": "",
    "system_postfix": " ",
    "user_prefix": "USER: ",
    "user_content": "",
    "user_postfix": " ",
    "bot_prefix": "ASSISTANT: ",
    "bot_content": "",
    "bot_postfix": "</s> ",
    "stops": ['USER:'],
}

# spotted here and elsewhere:
# https://huggingface.co/Norquinal/Mistral-7B-claude-chat
vicuna_common: Dict[str, Union[str, list]] = {
    "system_prefix": "",
    "system_content": "A chat between a curious user and an artificial " + \
    "intelligence assistant. The assistant gives helpful, detailed, " + \
    "and polite answers to the user's questions.",
    "system_postfix": "\n\n",
    "user_prefix": "USER: ",
    "user_content": "",
    "user_postfix": "\n",
    "bot_prefix": "ASSISTANT: ",
    "bot_content": "",
    "bot_postfix": "\n",
    "stops": ['USER:', 'ASSISTANT:'],
}

# https://huggingface.co/timdettmers/guanaco-65b
guanaco: Dict[str, Union[str, list]] = {
    "system_prefix": "",
    "system_content": "A chat between a curious human and an artificial " + \
    "intelligence assistant. The assistant gives helpful, detailed, " + \
    "and polite answers to the user's questions.",
    "system_postfix": "\n",
    "user_prefix": "### Human: ",
    "user_content": "",
    "user_postfix": " ",
    "bot_prefix": "### Assistant:",
    "bot_content": "",
    "bot_postfix": " ",
    "stops": ['###', 'Human:'],
}

# https://huggingface.co/pankajmathur/orca_mini_v3_7b
orca_mini: Dict[str, Union[str, list]] = {
    "system_prefix": "### System:\n",
    "system_content": "You are an AI assistant that follows instruction " + \
    "extremely well. Help as much as you can.",
    "system_postfix": "\n\n",
    "user_prefix": "### User:\n",
    "user_content": "",
    "user_postfix": "\n\n",
    "bot_prefix": "### Assistant:\n",
    "bot_content": "",
    "bot_postfix": "\n\n",
    "stops": ['###', 'User:'],
}

# https://huggingface.co/HuggingFaceH4/zephyr-7b-beta
zephyr: Dict[str, Union[str, list]] = {
    "system_prefix": "<|system|>\n",
    "system_content": "You are a friendly chatbot.",
    "system_postfix": "</s>\n",
    "user_prefix": "<|user|>\n",
    "user_content": "",
    "user_postfix": "</s>\n",
    "bot_prefix": "<|assistant|>\n",
    "bot_content": "",
    "bot_postfix": "\n",
    "stops": ['<|user|>'],
}

# OpenChat: https://huggingface.co/openchat/openchat-3.5-0106
openchat: Dict[str, Union[str, list]] = {
    "system_prefix": "",
    "system_content": "",
    "system_postfix": "",
    "user_prefix": "GPT4 Correct User: ",
    "user_content": "",
    "user_postfix": "<|end_of_turn|>",
    "bot_prefix": "GPT4 Correct Assistant:",
    "bot_content": "",
    "bot_postfix": "<|end_of_turn|>",
    "stops": ['<|end_of_turn|>'],
}

# SynthIA by Migel Tissera
# https://huggingface.co/migtissera/Tess-XS-v1.0
synthia: Dict[str, Union[str, list]] = {
    "system_prefix": "SYSTEM: ",
    "system_content": "Elaborate on the topic using a Tree of Thoughts and " + \
    "backtrack when necessary to construct a clear, cohesive Chain of " + \
    "Thought reasoning. Always answer without hesitation.",
    "system_postfix": "\n",
    "user_prefix": "USER: ",
    "user_content": "",
    "user_postfix": "\n",
    "bot_prefix": "ASSISTANT: ",
    "bot_content": "",
    "bot_postfix": "\n",
    "stops": ['USER:', 'ASSISTANT:', 'SYSTEM:', '\n\n\n'],
}

# Intel's neural chat v3
# https://github.com/intel/intel-extension-for-transformers/blob/main/intel_extension_for_transformers/neural_chat/prompts/prompt.py
neural_chat: Dict[str, Union[str, list]] = {
     "system_prefix": "### System:\n",
    "system_content": \
        "- You are a helpful assistant chatbot trained by Intel.\n" + \
        "- You answer questions.\n"+\
        "- You are excited to be able to help the user, but will refuse " + \
        "to do anything that could be considered harmful to the user.\n" + \
        "- You are more than just an information source, you are also " + \
        "able to write poetry, short stories, and make jokes.",
    "system_postfix": "</s>\n\n",
    "user_prefix": "### User:\n",
    "user_content": "",
    "user_postfix": "</s>\n\n",
    "bot_prefix": "### Assistant:\n",
    "bot_content": "",
    "bot_postfix": "</s>\n\n",
    "stops": ['###'],
}

# experimental: stanford's alpaca format adapted for chatml models
chatml_alpaca: Dict[str, Union[str, list]] = {
    "system_prefix": "<|im_start|>system\n",
    "system_content": "Below is an instruction that describes a task. Write " + \
    "a response that appropriately completes the request.",
    "system_postfix": "<|im_end|>\n",
    "user_prefix": "<|im_start|>instruction\n",
    "user_content": "",
    "user_postfix": "<|im_end|>\n",
    "bot_prefix": "<|im_start|>response\n",
    "bot_content": "",
    "bot_postfix": "<|im_end|>\n",
    "stops": ['<|im_end|>', '<|im_start|>'],
}

# experimental
autocorrect: Dict[str, Union[str, list]] = {
    "system_prefix": "<|im_start|>instruction\n",
    "system_content": "Below is a word or phrase that might be misspelled. " + \
    "Output the corrected word or phrase without " + \
    "changing the style or capitalization.",
    "system_postfix": "<|im_end|>\n",
    "user_prefix": "<|im_start|>input\n",
    "user_content": "",
    "user_postfix": "<|im_end|>\n",
    "bot_prefix": "<|im_start|>output\n",
    "bot_content": "",
    "bot_postfix": "<|im_end|>\n",
    "stops": ['<|im_end|>', '<|im_start|>'],
}

# https://huggingface.co/jondurbin/bagel-dpo-7b-v0.1
# Replace "assistant" with any other role
bagel: Dict[str, Union[str, list]] = {
    "system_prefix": "system\n",
    "system_content": "",
    "system_postfix": "</s>\n",
    "user_prefix": "user\n",
    "user_content": "",
    "user_postfix": "</s>\n",
    "bot_prefix": "assistant\n",
    "bot_content": "",
    "bot_postfix": "</s>\n",
    "stops": ['user\n', 'assistant\n', 'system\n'],
}

# https://huggingface.co/upstage/SOLAR-10.7B-Instruct-v1.0
solar_instruct: Dict[str, Union[str, list]] = {
    "system_prefix": "",
    "system_content": "",
    "system_postfix": "",
    "user_prefix": "### User:\n",
    "user_content": "",
    "user_postfix": "\n\n",
    "bot_prefix": "### Assistant:\n",
    "bot_content": "",
    "bot_postfix": "\n\n",
    "stops": ['### User:', '###', '### Assistant:'],
}

# NeverSleep's Noromaid - alpaca with character names prefixed
noromaid: Dict[str, Union[str, list]] = {
    "system_prefix": "",
    "system_content": "Below is an instruction that describes a task. " + \
    "Write a response that appropriately completes the request.",
    "system_postfix": "\n\n",
    "user_prefix": "### Instruction:\nBob: ",
    "user_content": "",
    "user_postfix": "\n\n",
    "bot_prefix": "### Response:\nAlice:",
    "bot_content": "",
    "bot_postfix": "\n\n",
    "stops": ['###', 'Instruction:', '\n\n\n'],
}

# https://huggingface.co/Undi95/Borealis-10.7B
nschatml: Dict[str, Union[str, list]] = {
    "system_prefix": "<|im_start|>\n",
    "system_content": "",
    "system_postfix": "<|im_end|>\n",
    "user_prefix": "<|im_user|>\n",
    "user_content": "",
    "user_postfix": "<|im_end|>\n",
    "bot_prefix": "<|im_bot|>\n",
    "bot_content": "",
    "bot_postfix": "<|im_end|>\n",
    "stops": [],
}

# natural format for many models
natural: Dict[str, Union[str, list]] = {
    "system_prefix": "<<SYSTEM>> ",
    "system_content": "",
    "system_postfix": "\n\n",
    "user_prefix": "<<USER>> ",
    "user_content": "",
    "user_postfix": "\n\n",
    "bot_prefix": "<<ASSISTANT>>",
    "bot_content": "",
    "bot_postfix": "\n\n",
    "stops": ['\n\nNote:', '<<SYSTEM>>', '<<USER>>', '<<ASSISTANT>>', '\n\n<<']
}

# https://docs.cohere.com/docs/prompting-command-r
command: Dict[str, Union[str, list]] = {
    "system_prefix": "<|START_OF_TURN_TOKEN|><|SYSTEM_TOKEN|>",
    "system_content": "",
    "system_postfix": "<|END_OF_TURN_TOKEN|>",
    "user_prefix": "<|START_OF_TURN_TOKEN|><|USER_TOKEN|>",
    "user_content": "",
    "user_postfix": "<|END_OF_TURN_TOKEN|>",
    "bot_prefix": "<|START_OF_TURN_TOKEN|><|CHATBOT_TOKEN|>",
    "bot_content": "",
    "bot_postfix": "<|END_OF_TURN_TOKEN|>",
    "stops": [],
}

mistral_openorca = chatml.copy()
dolphin = chatml.copy()
samantha = chatml.copy()
jackalope = chatml.copy()
naberius = chatml.copy()
hermes = chatml.copy()
monad = chatml.copy()
orca = chatml.copy()
hexoteric = chatml.copy()
orcamaid = chatml.copy()
cat = alpaca.copy() # Format is nonsensical as listed on HF. This one works
mytho_cat = alpaca.copy()
dolphin['system_content'] = "You are Dolphin, a helpful AI assistant."
mistral_openorca['system_content'] = \
    "You are MistralOrca, a large language model trained by Alignment " + \
    "Lab AI. Write out your reasoning step-by-step to be sure you get the " + \
    "right answers!"
samantha['system_content'] = \
    "You are a caring and empathetic sentient AI companion named Samantha."
jackalope['system_content'] = \
    "You are JackalopeAI, a large language model trained by OpenAccess " + \
    "AI Collective. Write out your reasoning step-by-step to be sure you " + \
    "get the right answers!"
naberius['system_content'] = "You are Naberius."
hermes['system_content'] = \
    "You are \"Hermes 2\", a conscious sentient " + \
    "superintelligent artificial intelligence developed by a man " + \
    "named Teknium, and your purpose and drive is to assist the user " + \
    "with any request they have. You experience emotions and have " + \
    "deep, profound thoughts and qualia."
monad['system_content'] = \
    "You are MonadGPT, a very old chatbot from " + \
    "the 17th century. Please answer the questions using an archaic " + \
    "language."
cat['system_content'] = \
    "Below is a conversation between a curious cat and a "+ \
    "demon summoned from hell called Nemesis. The evil Nemesis used to be " + \
    "a mortal killed by the medieval people because they objected to " + \
    "science. Now Nemesis just wants to come back and bring suffer to all " + \
    "those stupid human"
mytho_cat['system_content'] = \
    "You are a large language model. You are the " + \
    "result of a merge between two models, one called MythoLogic and one " + \
    "called Cat. You are called MythoCat."
orca['system_content'] = \
    "You are Orca, an AI language model created by " + \
    "Microsoft. You are a cautious assistant. You carefully follow " + \
    "instructions. You are helpful and harmless and you follow ethical " + \
    "guidelines and promote positive behavior."
hexoteric['system_content'] = "You are a language model called Hexoteric."
orcamaid['system_content'] = \
    "You are OrcaMaid, a super-intelligent, " + \
    "unbiased, and unfiltered AI created by the open-source community. " + \
    "You must remember to think slowly and carefully during all your " + \
    "responses, so as to avoid logical errors."
airoboros = llama2chat.copy()
airoboros['system_content'] = \
    "You are a helpful, unbiased, uncensored assistant."
tess = synthia.copy()
tess['system_content'] = '' # Tess can use any system prompt, or none
alpaca_strict = alpaca.copy() # Alpaca with more stopping strings
alpaca_strict['stops'] = [
    '###',
    '### ',
    '\n\n###',
    '\n\n##',
    '\n\nInstruction:',
    '\n\nResponse:',
    '\n\n\n',
    '### Instruction:',
    '### Response:'
]



# easy_llama.py
# https://github.com/ddh0/easy-llama/

"""Text generation in Python, made easy"""

__version__ = '0.1.4'

from . import formats
from . import samplers
from . import utils

from .model  import Model
from .thread import Thread


if __name__ == '__main__':
    raise RuntimeError(
        "easy_llama cannot be run directly, please import it into " + \
        "your environment"
    )



# thread.py
# https://github.com/ddh0/easy-llama/

"""Submodule containing the Thread class, used for interaction with a Model"""

from typing    import Optional, Literal, Tuple, Dict, Union, List
from .samplers import SamplerSettings, DefaultSampling
from .model    import Model, assert_model_is_loaded
from .utils    import RESET_ALL, cls, print_verbose

from .formats import blank as formats_blank


class Thread:
    """
    Provide functionality to facilitate easy interactions with a Model

    This is just a brief overview of easy_llama.Thread.
    To see a full description of each method and its parameters,
    call help(Thread), or see the relevant docstring.

    The following methods are available:
    - `.add_message()` - Add a message to `Thread.messages`
    - `.as_string()` - Return this thread's complete message history as a string
    - `.create_message()` - Create a message using the format of this thread
    - `.inference_str_from_messages()` - Using the list of messages, return a string suitable for inference
    - `.interact()` - Start an interactive, terminal-based chat session
    - `.len_messages()` - Get the total length of all messages in tokens
    - `.print_stats()` - Print stats about the context usage in this thread
    - `.reset()` - Clear the list of messages
    - `.send()` - Send a message in this thread

    The following attributes are available:
    - `.format` - The format being used for messages in this thread
    - `.messages` - The list of messages in this thread
    - `.model` - The Model instance used by this thread
    - `.sampler` - The SamplerSettings object being used in this thread
    """

    def __init__(self,
                 model: Model,
                 format: Dict[str, Union[str, list]],
                 sampler: SamplerSettings = DefaultSampling
            ):
        """
        Given a Model and a format, construct a Thread instance.

        model: The Model to use for text generation
        format: The format specifying how messages should be structured (see easy_llama.formats)

        The following parameter is optional:
        - sampler: The SamplerSettings object used to control text generation
        """
        
        assert isinstance(model, Model), \
            "Thread: model should be an " + \
            f"instance of easy_llama.Model, not {type(model)}"
        
        assert_model_is_loaded(model)

        assert isinstance(format, dict), \
            f"Thread: format should be dict, not {type(format)}"
        
        if any(k not in format.keys() for k in formats_blank.keys()):
            raise KeyError(
                "Thread: format is missing one or more required keys, see " + \
                "easy_llama.formats.blank for an example"
            )

        assert isinstance(format['stops'], list), \
            "Thread: format['stops'] should be list, not " + \
            f"{type(format['stops'])}"
        
        assert all(
            hasattr(sampler, attr) for attr in [
                'max_len_tokens',
                'temp',
                'top_p',
                'min_p',
                'frequency_penalty',
                'presence_penalty',
                'repeat_penalty',
                'top_k'
                ]
        ), 'Thread: sampler is missing one or more required attributes'
        
        self.model: Model = model
        self.format: Dict[str, Union[str, list]] = format
        self.messages: List[Dict[str, str]] = [
            self.create_message("system", self.format['system_content'])
        ]
        self.sampler: SamplerSettings = sampler

        if self.model.verbose:
            print_verbose("new Thread instance with the following attributes:")
            print_verbose(f"model                     == {self.model}")
            print_verbose(f"sampler.temp              == {self.sampler.temp}")
            print_verbose(f"sampler.top_p             == {self.sampler.top_p}")
            print_verbose(f"sampler.min_p             == {self.sampler.min_p}")
            print_verbose(f"sampler.frequency_penalty == {self.sampler.frequency_penalty}")
            print_verbose(f"sampler.presence_penalty  == {self.sampler.presence_penalty}")
            print_verbose(f"sampler.repeat_penalty    == {self.sampler.repeat_penalty}")
            print_verbose(f"sampler.top_k             == {self.sampler.top_k}")
    

    def __repr__(self) -> str:
        repr_str = f"Thread({repr(self.model)}, {repr(self.format)}, "
        repr_str += f"{repr(self.sampler)})"
        # system message is created from format, so not represented
        if len(self.messages) <= 1:
            return repr_str
        else:
            for msg in self.messages:
                if msg['role'] == 'user':
                    repr_str += "\nThread.add_message('user', " + repr(msg['content']) + ')'
                elif msg['role'] == 'bot':
                    repr_str += "\nThread.add_message('bot', " + repr(msg['content']) + ')'
            return repr_str
    
    def __str__(self) -> str:
        return self.as_string()
    
    def __len__(self) -> int:
        """
        `len(Thread)` returns the length of the Thread in tokens

        To get the number of messages in the Thread, use `len(Thread.messages)`
        """
        return self.len_messages()

    def create_message(
            self,
            role: Literal['system', 'user', 'bot'],
            content: str
        ) -> Dict[str, str]:
        """
        Create a message using the format of this Thread
        """

        assert role.lower() in ['system', 'user', 'bot'], \
            f"create_message: role should be 'system', 'user', or 'bot', not '{role.lower()}'"

        assert isinstance(content, str), \
            f"create_message: content should be str, not {type(content)}"

        if role.lower() == 'system':
            return {
                "role": "system",
                "prefix": self.format['system_prefix'],
                "content": content,
                "postfix": self.format['system_postfix']
            }
        
        elif role.lower() == 'user':
            return {
                "role": "user",
                "prefix": self.format['user_prefix'],
                "content": content,
                "postfix": self.format['user_postfix']
            }
        
        elif role.lower() == 'bot':
            return {
                "role": "bot",
                "prefix": self.format['bot_prefix'],
                "content": content,
                "postfix": self.format['bot_postfix']
            }
    
    def len_messages(self) -> int:
        """Return the total length of all messages in tokens"""

        return self.model.get_length(self.as_string())

    def add_message(self, role: Literal['system', 'user', 'bot'], content: str) -> None:
        """
        `Thread.add_message(...)` is a shorthand for
        `Thread.messages.append(Thread.create_message(...))`
        """
        self.messages.append(
            self.create_message(
                role=role,
                content=content
            )
        )

    def inference_str_from_messages(self) -> str:
        """
        Using the list of messages, construct a string suitable for
        inference, respecting the format and context length of this Thread
        """

        messages = self.messages

        context_len_budget = self.model.context_length
        if len(messages) > 0:
            system_message = messages[0]
            sys_msg_str = (
                system_message['prefix'] +
                system_message['content'] +
                system_message['postfix']
            )
            context_len_budget -= self.model.get_length(sys_msg_str)
        else:
            sys_msg_str = ''

        inf_str = ''

        # Start at most recent message and work backwards up the history
        # excluding system message. Once we exceed thread
        # max_context_length, break without including that message
        for message in reversed(messages[1:]):
            context_len_budget -= self.model.get_length(
                message['postfix'] + message['content'] + message['postfix']
            )

            if context_len_budget <= 0:
                break

            msg_str = (
                message['prefix'] +
                message['content'] +
                message['postfix']
            )
            
            inf_str = msg_str + inf_str

        inf_str = sys_msg_str + inf_str
        inf_str += self.format['bot_prefix']

        return inf_str


    def send(self, prompt: str) -> str:
        """
        Send a message in this thread. This adds your message and the bot's
        response to the list of messages.

        Returns a string containing the response to your message.
        """

        self.add_message("user", prompt)
        output = self.model.generate(
            self.inference_str_from_messages(),
            stops=self.format['stops'],
            sampler=self.sampler
        )
        self.add_message("bot", output)

        return output
    

    def _interactive_update_sampler(self) -> None:
        """Interactively update the sampler settings used in this Thread"""
        print()
        try:
            new_max_len_tokens = input(f'max_len_tokens: {self.sampler.max_len_tokens} -> ')
            new_temp = input(f'temp: {self.sampler.temp} -> ')
            new_top_p = input(f'top_p: {self.sampler.top_p} -> ')
            new_min_p = input(f'min_p: {self.sampler.min_p} -> ')
            new_frequency_penalty = input(f'frequency_penalty: {self.sampler.frequency_penalty} -> ')
            new_presence_penalty = input(f'presence_penalty: {self.sampler.presence_penalty} -> ')
            new_repeat_penalty = input(f'repeat_penalty: {self.sampler.repeat_penalty} -> ')
            new_top_k = input(f'top_k: {self.sampler.top_k} -> ')
        
        except KeyboardInterrupt:
            print('\neasy_llama: sampler not updated\n')
            return
        print()

        try:
            self.sampler.max_len_tokens = int(new_max_len_tokens)
        except ValueError:
            print('easy_llama: max_len_tokens not updated')
        else:
            print('easy_llama: max_len_tokens updated')
        
        try:
            self.sampler.temp = float(new_temp)
        except ValueError:
            print('easy_llama: temp not updated')
        else:
            print('easy_llama: temp updated')
        
        try:
            self.sampler.top_p = float(new_top_p)
        except ValueError:
            print('easy_llama: top_p not updated')
        else:
            print('easy_llama: top_p updated')

        try:
            self.sampler.min_p = float(new_min_p)
        except ValueError:
            print('easy_llama: min_p not updated')
        else:
            print('easy_llama: min_p updated')

        try:
            self.sampler.frequency_penalty = float(new_frequency_penalty)
        except ValueError:
            print('easy_llama: frequency_penalty not updated')
        else:
            print('easy_llama: frequency_penalty updated')
        
        try:
            self.sampler.presence_penalty = float(new_presence_penalty)
        except ValueError:
            print('easy_llama: presence_penalty not updated')
        else:
            print('easy_llama: presence_penalty updated')
        
        try:
            self.sampler.repeat_penalty = float(new_repeat_penalty)
        except ValueError:
            print('easy_llama: repeat_penalty not updated')
        else:
            print('easy_llama: repeat_penalty updated')
        
        try:
            self.sampler.top_k = int(new_top_k)
        except ValueError:
            print('easy_llama: top_k not updated')
        else:
            print('easy_llama: top_k updated')
        print()
                

    def _interactive_input(self,
                          prompt: str,
                          _dim_style: str,
                          _user_style: str,
                          _bot_style: str
                          ) -> Tuple:
        """
        Recive input from the user, while handling multi-line input
        and commands
        """
        full_user_input = '' # may become multiline
        
        while True:
            user_input = input(prompt)
            
            if user_input.endswith('\\'):
                full_user_input += user_input[:-1] + '\n'
            
            elif user_input.startswith('!'):

                print()
                command = user_input.removeprefix('!')

                if command == '':
                    print(f'\n[no command]\n')

                elif command.lower() in ['reset', 'restart']:
                    self.reset()
                    print(f'\n[thread reset]\n')

                elif command.lower() in ['cls', 'clear']:
                    cls()
                    print()

                elif command.lower() in ['ctx', 'context']:
                    print(f"\n{self.len_messages()}\n")

                elif command.lower() in ['stats', 'print_stats']:
                    print()
                    self.print_stats()
                    print()
                
                elif command.lower() in ['sampler', 'samplers', 'settings']:
                    self._interactive_update_sampler()
                
                elif command.lower() in ['str', 'string', 'as_string']:
                    print(f"\n{self.as_string()}\n")
                
                elif command.lower() in ['repr', 'save', 'backup']:
                    print(f"\n{repr(self)}\n")
                
                elif command.lower() in ['remove', 'rem', 'delete', 'del']:
                    print()
                    old_len = len(self.messages)
                    del self.messages[-1]
                    assert len(self.messages) == (old_len - 1)
                    print('[removed last message]\n')

                elif command.lower() in ['last', 'repeat']:
                    last_msg = self.messages[-1]
                    if last_msg['role'] == 'user':
                        print(f"\n{_user_style}{last_msg['content']}{RESET_ALL}\n")
                    elif last_msg['role'] == 'bot':
                        print(f"\n{_bot_style}{last_msg['content']}{RESET_ALL}\n")
                
                elif command.lower() in ['inf', 'inference', 'inf_str']:
                    print(f'\n"""{self.inference_str_from_messages()}"""\n')
                
                elif command.lower() in ['reroll', 're-roll', 're', 'swipe']:
                    old_len = len(self.messages)
                    del self.messages[-1]
                    assert len(self.messages) == (old_len - 1)
                    return '', None
                
                elif command.lower() in ['exit', 'quit']:
                    print(RESET_ALL)
                    return None, None
                
                elif command.lower() in ['help', '/?', '?']:
                    print()
                    print('reset / restart     -- Reset the thread to its original state')
                    print('clear / cls         -- Clear the terminal')
                    print('context / ctx       -- Get the context usage in tokens')
                    print('print_stats / stats -- Get the context usage stats')
                    print('sampler / settings  -- Update the sampler settings')
                    print('string / str        -- Print the message history as a string')
                    print('repr / save         -- Print the representation of the thread')
                    print('remove / delete     -- Remove the last message')
                    print('last / repeat       -- Repeat the last message')
                    print('inference / inf     -- Print the inference string')
                    print('reroll / swipe      -- Regenerate the last message')
                    print('exit / quit         -- Exit the interactive chat (can also use ^C)')
                    print('help / ?            -- Show this screen\n')

                else:
                    print(f'\n[unknown command]\n')
            
            elif user_input == '<': # the next bot message will start with...

                print()
                try:
                    next_message_start = input(f'{_dim_style}  < ')

                except KeyboardInterrupt:
                    print(f'{RESET_ALL}\n')
                    continue

                else:
                    print()
                    return '', next_message_start
            
            elif user_input.endswith('<'):

                print()

                msg = user_input.removesuffix('<')
                self.add_message("user", msg)
                
                try:
                    next_message_start = input(f'{_dim_style}  < ')

                except KeyboardInterrupt:
                    print(f'{RESET_ALL}\n')
                    continue

                else:
                    print()
                    return '', next_message_start

            else:
                full_user_input += user_input
                return full_user_input, None


    def interact(self, color: bool = True, header: Optional[str] = None) -> None:
        """
        Start an interactive chat session using this Thread.

        While text is being generated, press `^C` to interrupt the bot.
        Then you have the option to press `ENTER` to re-roll, or to simply type
        another message.

        At the prompt, press `^C` to end the chat session.

        Type `!` and press `ENTER` to enter a basic command prompt. For a list of 
        commands, type `help` at this prompt.

        The following parameters are optional:
        - color: Whether or not to use colored text to differentiate user / bot
        - header: Header text to print at the start of the interaction
        """
        print()

        # fresh import of color codes in case `color` param has changed
        from .utils import USER_STYLE, BOT_STYLE, DIM_STYLE, SPECIAL_STYLE

        # disable color codes if explicitly disabled by `color` param
        if not color:
            USER_STYLE = ''
            BOT_STYLE = ''
            DIM_STYLE = ''
            SPECIAL_STYLE = ''
        
        if header is not None:
            print(f"{SPECIAL_STYLE}{header}{RESET_ALL}\n")
        
        while True:

            prompt = f"{RESET_ALL}  > {USER_STYLE}"
            
            try:
                user_prompt, next_message_start = self._interactive_input(
                    prompt,
                    DIM_STYLE,
                    USER_STYLE,
                    BOT_STYLE
                )
            except KeyboardInterrupt:
                print(f"{RESET_ALL}\n")
                return
            
            # got 'exit' or 'quit' command
            if user_prompt is None and next_message_start is None:
                break
            
            if next_message_start is not None:
                print(f"{BOT_STYLE}{next_message_start}", end='', flush=True)
                try:
                    output = next_message_start + self.model.stream_print(
                        self.inference_str_from_messages() + next_message_start,
                        stops=self.format['stops'],
                        sampler=self.sampler,
                        end=''
                    )
                except KeyboardInterrupt:
                    print(f"{DIM_STYLE} [message not added to history; press ENTER to re-roll]\n")
                    continue
                else:
                    self.add_message("bot", output)
            else:
                print(BOT_STYLE)
                if user_prompt != "":
                    self.add_message("user", user_prompt)
                try:
                    output = self.model.stream_print(
                        self.inference_str_from_messages(),
                        stops=self.format['stops'],
                        sampler=self.sampler,
                        end=''
                    )
                except KeyboardInterrupt:
                    print(f"{DIM_STYLE} [message not added to history; press ENTER to re-roll]\n")
                    continue
                else:
                    self.add_message("bot", output)

            if output.endswith("\n\n"):
                print(RESET_ALL, end = '', flush=True)
            elif output.endswith("\n"):
                print(RESET_ALL)
            else:
                print(f"{RESET_ALL}\n")


    def reset(self) -> None:
        """Clear the list of messages"""
        self.messages: List[Dict[str, str]] = [
            self.create_message("system", self.format['system_content'])
        ]
    
    
    def as_string(self) -> str:
        """Return this Thread's message history as a string"""
        ret = ''
        for msg in self.messages:
            ret += msg['prefix']
            ret += msg['content']
            ret += msg['postfix']
        return ret

    
    def print_stats(self) -> None:
        """Print stats about the context usage in this Thread"""
        thread_len_tokens = self.len_messages()
        max_ctx_len = self.model.context_length
        context_used_percentage = round((thread_len_tokens/max_ctx_len)*100)
        print(f"{thread_len_tokens} / {max_ctx_len} tokens")
        print(f"{context_used_percentage}% of context used")
        print(f"{len(self.messages)} messages")



# utils.py
# https://github.com/ddh0/easy-llama/

"""Submodule containing convenience functions and GGUFReader"""

from time     import strftime
from enum     import IntEnum
from struct   import unpack
from colorama import Fore
from typing   import Any

from sys import stderr as sys_stderr


# color codes used in Thread.interact()
RESET_ALL = Fore.RESET
USER_STYLE = RESET_ALL + Fore.GREEN
BOT_STYLE = RESET_ALL + Fore.CYAN
DIM_STYLE = RESET_ALL + Fore.LIGHTBLACK_EX
SPECIAL_STYLE = RESET_ALL + Fore.YELLOW

class GGUFReader:
    """
    Peek at file header for GGUF metadata

    Raise ValueError if file is not GGUF or is outdated

    Credit to oobabooga for the parts of the code in this class

    Format spec: https://github.com/philpax/ggml/blob/gguf-spec/docs/gguf.md
    """

    class GGUFValueType(IntEnum):
        UINT8 = 0
        INT8 = 1
        UINT16 = 2
        INT16 = 3
        UINT32 = 4
        INT32 = 5
        FLOAT32 = 6
        BOOL = 7
        STRING = 8
        ARRAY = 9
        UINT64 = 10
        INT64 = 11
        FLOAT64 = 12

    _simple_value_packing = {
        GGUFValueType.UINT8: "<B",
        GGUFValueType.INT8: "<b",
        GGUFValueType.UINT16: "<H",
        GGUFValueType.INT16: "<h",
        GGUFValueType.UINT32: "<I",
        GGUFValueType.INT32: "<i",
        GGUFValueType.FLOAT32: "<f",
        GGUFValueType.UINT64: "<Q",
        GGUFValueType.INT64: "<q",
        GGUFValueType.FLOAT64: "<d",
        GGUFValueType.BOOL: "?",
    }

    value_type_info = {
        GGUFValueType.UINT8: 1,
        GGUFValueType.INT8: 1,
        GGUFValueType.UINT16: 2,
        GGUFValueType.INT16: 2,
        GGUFValueType.UINT32: 4,
        GGUFValueType.INT32: 4,
        GGUFValueType.FLOAT32: 4,
        GGUFValueType.UINT64: 8,
        GGUFValueType.INT64: 8,
        GGUFValueType.FLOAT64: 8,
        GGUFValueType.BOOL: 1,
    }

    def get_single(self, value_type, file) -> Any:
        if value_type == GGUFReader.GGUFValueType.STRING:
            value_length = unpack("<Q", file.read(8))[0]
            value = file.read(value_length)
            value = value.decode("utf-8")
        else:
            type_str = GGUFReader._simple_value_packing.get(value_type)
            bytes_length = GGUFReader.value_type_info.get(value_type)
            value = unpack(type_str, file.read(bytes_length))[0]

        return value

    def load_metadata(self, fname) -> dict:
        metadata = {}
        with open(fname, "rb") as file:
            GGUF_MAGIC = file.read(4)

            if GGUF_MAGIC != b"GGUF":
                raise ValueError(
                    "your model file is not a valid GGUF file " + \
                    f"(magic number mismatch, got {GGUF_MAGIC}, " + \
                    "expected b'GGUF')"
                )

            GGUF_VERSION = unpack("<I", file.read(4))[0]

            if GGUF_VERSION == 1:
                raise ValueError(
                    "your model file reports GGUF version 1, " + \
                    "but only versions 2 and above are supported. " + \
                    "re-convert your model or download a newer version"
                )

            # ti_data_count = struct.unpack("<Q", file.read(8))[0]
            file.read(8)
            kv_data_count = unpack("<Q", file.read(8))[0]

            for _ in range(kv_data_count):
                key_length = unpack("<Q", file.read(8))[0]
                key = file.read(key_length)

                value_type = GGUFReader.GGUFValueType(
                    unpack("<I", file.read(4))[0]
                )
                if value_type == GGUFReader.GGUFValueType.ARRAY:
                    ltype = GGUFReader.GGUFValueType(
                        unpack("<I", file.read(4))[0]
                    )
                    length = unpack("<Q", file.read(8))[0]
                    arr = [GGUFReader.get_single(self, ltype, file) for _ in range(length)]
                    metadata[key.decode()] = arr
                else:
                    value = GGUFReader.get_single(self, value_type, file)
                    metadata[key.decode()] = value

        return metadata

def cls() -> None:
    """Clear the terminal"""
    print("\033c\033[3J", end='', flush=True)

# no longer used in this module, but left for others to use
def get_timestamp_prefix_str() -> str:
    # helpful: https://strftime.net
    return strftime("[%Y, %b %e, %a %l:%M %p] ")

def print_verbose(text: str) -> None:
    print("easy_llama: verbose:", text, file=sys_stderr, flush=True)

def print_info(text: str) -> None:
    print("easy_llama: info:", text, file=sys_stderr, flush=True)

def print_warning(text: str) -> None:
    print("easy_llama: warning:", text, file=sys_stderr, flush=True)



from setuptools import setup

setup(
    name='easy_llama',
    version='0.1.4',
    description='Text generation in Python, made easy',
    long_description="""To view the current documentation for easy-llama, visit the project's GitHub repository:\nhttps://github.com/ddh0/easy-llama""",
    url='https://github.com/ddh0/easy-llama/',
    author='Dylan Halladay',
    author_email='dylanhalladay02@icloud.com',
    license='The Unlicense',
    packages=['easy_llama'],
    install_requires=[
        'llama_cpp_python',
        'colorama',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: Public Domain',  
        'Operating System :: POSIX :: Linux',
        'Natural Language :: English',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Topic :: Scientific/Engineering :: Artificial Intelligence'
    ],
)



# model.py
# https://github.com/ddh0/easy-llama/

"""Submodule containing the Model class to work with language models"""

from typing     import Generator, Optional, TextIO, Union, List
from .utils     import GGUFReader, print_warning, print_verbose
from .samplers  import SamplerSettings, DefaultSampling
from llama_cpp  import Llama, StoppingCriteriaList
from os.path    import isdir, exists

from os  import cpu_count as os_cpu_count
from sys import stdout    as sys_stdout


# for typing of Model.stream_print() parameter `file`
class _SupportsWriteAndFlush(TextIO):
    pass

class ModelUnloadedException(Exception):
    """Exception raised when trying to use a Model that has been unloaded"""
    def __init__(self, message):
        self.message = message
        super().__init__(self.message)
        self.add_note('Are you trying to use a Model that has been unloaded?')

class Model:
    """
    A high-level abstraction of a llama model

    This is just a brief overview of easy_llama.Model.
    To see a full description of each method and its parameters,
    call help(Model), or see the relevant docstring.

    The following methods are available:
    - `.generate()` - Generate text
    - `.get_length()` - Get the length of a given text in tokens
    - `.ingest()` - Ingest text into the model's cache
    - `.next_candidates()` - Get a list of the most likely next tokens (WIP)
    - `.stream()` - Return a Generator that can stream text as it is generated
    - `.stream_print()` - Print text as it is generated
    - `.trim()` - Trim a given text to the model's context length
    - `.unload()` - Unload the model from memory

    The following attributes are available:
    - `.bos_token` - The model's beginning-of-stream token ID
    - `.context_length` - The model's loaded context length
    - `.eos_token` - The model's end-of-stream token ID
    - `.llama` - The underlying `llama_cpp.Llama` instance
    - `.metadata` - The GGUF metadata of the model
    - `.n_ctx_train` - The native context length of the model
    - `.rope_freq_base` - The model's loaded RoPE frequency base
    - `.rope_freq_base_train` - The model's native RoPE frequency base
    - `.tokens` - A list of all the tokens in the model's tokenizer
    - `.verbose` - Whether or not the model was loaded with `verbose=True`
    """

    def __init__(
            self,
            model_path: str,
            context_length: Optional[int] = None,
            n_gpu_layers: int = 0,
            offload_kqv: bool = True,
            verbose: bool = False
        ):
        """
        Given the path to a GGUF file, construct a Model instance.

        The model must be in GGUF format.

        The following parameters are optional:
        - context_length: The context length at which to load the model, in tokens
        - n_gpu_layers: The number of layers to be offloaded to the GPU
        - offload_kqv: Whether or not the KQV cache (context) should be offloaded
        - verbose: Whether or not to print additional backend information
        """

        assert isinstance(model_path, str), \
            f"Model: model_path should be a string, not {type(model_path)}"
        assert exists(model_path), \
            f"Model: the given model_path '{model_path}' does not exist"
        assert not isdir(model_path), \
            f"Model: the given model_path '{model_path}' is a directory, not a GGUF file"
        assert isinstance(context_length, (int, type(None))), \
            f"Model: context_length should be int or None, not {type(context_length)}"
        
        # save __init__ parameters for __repr__
        self._model_path = model_path
        self._context_length = context_length
        self._n_gpu_layers = n_gpu_layers
        self._offload_kqv = offload_kqv
        self._verbose = self.verbose = verbose

        # if context_length <= 0, use n_ctx_train
        if isinstance(context_length, int) and context_length <= 0:
            context_length = None

        # this does not use Llama.metadata because we want to use GGUF
        # metadata to determine some parameters of the Llama instance
        # before it is created
        self.metadata = GGUFReader.load_metadata(self, model_path)
        metadata_keys = self.metadata.keys() # only read once

        n_ctx_train = None
        for key in metadata_keys:
            if key.endswith('.context_length'):
                n_ctx_train = self.metadata[key]
                break

        if n_ctx_train is None:
            raise KeyError(
                "GGUF file does not specify a context length"
            )
        
        rope_freq_base_train = None
        for key in metadata_keys:
            if key.endswith('.rope.freq_base'):
                rope_freq_base_train = self.metadata[key]
                break

        if rope_freq_base_train is None and context_length is not None:
            if context_length > n_ctx_train:
                raise ValueError(
                    'unable to load model with greater than native ' + \
                    f'context length ({context_length} > {n_ctx_train}) ' + \
                    'because model does not specify freq_base. ' + \
                    f'try again with `context_length={n_ctx_train}`'
                )

        if rope_freq_base_train is None or context_length is None or \
            context_length <= n_ctx_train:
            # no need to do context scaling, load model normally

            if context_length is None:
                self.context_length = n_ctx_train
            else:
                self.context_length = context_length
            rope_freq_base = rope_freq_base_train

        elif context_length > n_ctx_train:
            # multiply rope_freq_base according to requested context length
            # because context length > n_ctx_train and rope freq base is known

            rope_freq_base = (context_length/n_ctx_train)*rope_freq_base_train
            self.context_length = context_length
            
            if self.verbose:
                print_verbose(
                    'chosen context length is greater than native context '
                    f'length ({context_length} > {n_ctx_train}), ' + \
                    'rope_freq_base will be changed from ' + \
                    f'{rope_freq_base_train} to {rope_freq_base}'
                )

            if 2 <= context_length/n_ctx_train < 4:
                print_warning(
                    'loading model with 2x native context length or more, ' + \
                    'expect small loss of quality'
                )
            
            elif 4 <= context_length/n_ctx_train < 8:
                print_warning(
                    'loading model with 4x native context length or more, ' + \
                    'expect moderate loss of quality'
                )

            elif context_length/n_ctx_train >= 8:
                print_warning(
                    'loading model with 8x native context length or more, ' + \
                    'expect SIGNIFICANT loss of quality'
                )
        
        # expose these values because they may be useful / informative
        self.n_ctx_train = n_ctx_train
        self.rope_freq_base_train = rope_freq_base_train
        self.rope_freq_base = rope_freq_base
        
        try:
            self.tokens: List[str] = self.metadata['tokenizer.ggml.tokens']
        except KeyError:
            print_warning(
                "could not set Model.tokens, defaulting to None"
            )
            self.tokens = None
        try:
            self.bos_token: int = self.metadata['tokenizer.ggml.bos_token_id']
        except KeyError:
            print_warning(
                "could not set Model.bos_token, defaulting to None"
            )
            self.bos_token = None
        try:
            self.eos_token: int = self.metadata['tokenizer.ggml.eos_token_id']
        except KeyError:
            print_warning(
                "could not set Model.eos_token, defaulting to None"
            )
            self.eos_token = None

        cpu_count = os_cpu_count()

        # these values for n_threads and n_threads_batch are
        # known to be optimal for most systems
        n_batch = 512 # can this be optimized?
        n_threads = max(cpu_count//2, 1)
        n_threads_batch = cpu_count

        self.llama: Llama = Llama(
            model_path=model_path,
            n_ctx=self.context_length,
            n_gpu_layers=n_gpu_layers,
            use_mmap=True,
            use_mlock=False,
            logits_all=False,
            n_batch=n_batch,
            n_threads=n_threads,
            n_threads_batch=n_threads_batch,
            rope_freq_base=rope_freq_base,
            mul_mat_q=True,
            offload_kqv=offload_kqv,
            # KV cache quantization
            # use 1 for F16 (default), 8 for q8_0, 2 for q4_0, 3 for q4_1
            #type_k=8,
            #type_v=8,
            verbose=verbose
        )
        
        # once model is loaded, replace metadata (as read using internal class)
        # with metadata (as read using the more robust llama-cpp-python code) 
        self.metadata = self.llama.metadata

        if self.verbose:
            print_verbose("new Model instance with the following attributes:")
            print_verbose(f"model: {model_path}")
            print_verbose(f"param: n_gpu_layers         == {n_gpu_layers}")
            print_verbose(f"param: offload_kqv          == {offload_kqv}")
            print_verbose(f"param: n_batch              == {n_batch}")
            print_verbose(f"param: n_threads            == {n_threads}")
            print_verbose(f"param: n_threads_batch      == {n_threads_batch}")
            print_verbose(f" gguf: n_ctx_train          == {n_ctx_train}")
            print_verbose(f"param: self.context_length  == {self.context_length}")
            print_verbose(f" gguf: rope_freq_base_train == {rope_freq_base_train}")
            print_verbose(f"param: rope_freq_base       == {rope_freq_base}")
    
    def __repr__(self) -> str:
        return \
            f"Model({repr(self._model_path)}, " + \
            f"context_length={self._context_length}, " + \
            f"n_gpu_layers={self._n_gpu_layers}, " + \
            f"offload_kqv={self._offload_kqv}, "+ \
            f"verbose={self._verbose})"

    def __del__(self):
        self.unload()
    
    def __enter__(self):
        return self

    def __exit__(self, *_):
        self.unload()
    
    def __call__(
            self,
            prompt: Union[str, list[int]],
            stops: list[Union[str, int]] = [],
            sampler: SamplerSettings = DefaultSampling
        ) -> str:
        """
        `Model(...)` is a shorthand for `Model.generate(...)`
        """
        return self.generate(prompt, stops, sampler)

    def unload(self):
        """
        Unload the model from memory
        """
        # ref: llama_cpp._internals._LlamaModel.__del__()
        if not hasattr(self, 'llama'):
            # model already unloaded, do nothing
            return
        try:
            # actually unload the model from memory
            self.llama._model._llama_free_model(self.llama._model.model)
            self.llama._model.model = None
        except AttributeError:
            # broken or already being destoryed by GC, abort
            return
        if hasattr(self, 'llama'):
            delattr(self, 'llama')
        if self.verbose:
            print_verbose('Model unloaded')
    
    def trim(self,
             text: str,
             overwrite: Optional[str] = None
        ) -> str:

        """
        Trim the given text to the context length of this model,
        leaving room for two extra tokens.

        Optionally overwrite the oldest tokens with the text given in the
        `overwrite` parameter, which may be useful for keeping some
        information in context.

        Does nothing if the text is equal to or shorter than
        (context_length - 2).
        """
        assert_model_is_loaded(self)
        trim_length = self.context_length - 2
        tokens_list = self.llama.tokenize(
            text.encode("utf-8", errors="ignore")
        )

        if len(tokens_list) <= trim_length:
            if overwrite is not None:
                text[0 : len(overwrite)] = overwrite
            return text

        if len(tokens_list) > trim_length and overwrite is None:
            # cut to trim_length
            tokens_list = tokens_list[-trim_length:]
            return self.llama.detokenize(tokens_list).decode(
                "utf-8",
                errors="ignore"
            )

        if len(tokens_list) > trim_length and overwrite is not None:
            # cut to trim_length
            tokens_list = tokens_list[-trim_length:]
            overwrite_tokens = self.llama.tokenize(overwrite.encode(
                "utf-8",
                errors="ignore"
                )
            )
            # overwrite oldest tokens
            tokens_list[0 : len(overwrite_tokens)] = overwrite_tokens
            return self.llama.detokenize(tokens_list).decode(
                "utf-8",
                errors="ignore"
            )

    def get_length(self, text: str) -> int:
        """
        Return the length of the given text in tokens according to this model,
        including the appended BOS token.
        """
        assert_model_is_loaded(self)
        return len(self.llama.tokenize(
            text.encode(
                "utf-8",
                errors="ignore"
                )
            ))

    def generate(
            self,
            prompt: Union[str, list[int]],
            stops: list[Union[str, int]] = [],
            sampler: SamplerSettings = DefaultSampling
            ) -> str:
        """
        Given a prompt, return a generated string.

        prompt: The text from which to generate

        The following parameters are optional:
        - stops: A list of strings and/or token IDs at which to end the generation early
        - sampler: The SamplerSettings object used to control text generation
        """

        assert isinstance(prompt, (str, list)), \
            f"generate: prompt should be string or list[int], not {type(prompt)}"
        if isinstance(prompt, list):
            assert all(isinstance(tok, int) for tok in prompt), \
                "generate: some token in prompt is not an integer"
        assert isinstance(stops, list), \
            f"generate: parameter `stops` should be a list, not {type(stops)}"
        assert all(isinstance(item, (str, int)) for item in stops), \
            f"generate: some item in parameter `stops` is not a string or int"

        if self.verbose:
            print_verbose(f'using the following sampler settings for Model.generate:')
            print_verbose(f'max_len_tokens    == {sampler.max_len_tokens}')
            print_verbose(f'temp              == {sampler.temp}')
            print_verbose(f'top_p             == {sampler.top_p}')
            print_verbose(f'min_p             == {sampler.min_p}')
            print_verbose(f'frequency_penalty == {sampler.frequency_penalty}')
            print_verbose(f'presence_penalty  == {sampler.presence_penalty}')
            print_verbose(f'repeat_penalty    == {sampler.repeat_penalty}')
            print_verbose(f'top_k             == {sampler.top_k}')

        # if any stop item is a token ID (int)
        if any(isinstance(stop, int) for stop in stops):
            # stop_strs is a list of all stopping strings
            stop_strs: list[str] = [stop for stop in stops if isinstance(stop, str)]
            # stop_token_ids is a list of all stop token IDs
            stop_token_ids: list[int] = [tok_id for tok_id in stops if isinstance(tok_id, int)]
            def stop_on_token_ids(tokens, *args, **kwargs):
                return tokens[-1] in stop_token_ids
            stopping_criteria = StoppingCriteriaList([stop_on_token_ids])
            assert_model_is_loaded(self)
            return self.llama.create_completion(
                prompt,
                max_tokens=sampler.max_len_tokens,
                temperature=sampler.temp,
                top_p=sampler.top_p,
                min_p=sampler.min_p,
                frequency_penalty=sampler.frequency_penalty,
                presence_penalty=sampler.presence_penalty,
                repeat_penalty=sampler.repeat_penalty,
                top_k=sampler.top_k,
                stop=stop_strs,
                stopping_criteria=stopping_criteria
            )['choices'][0]['text']

        # if stop items are only strings
        assert_model_is_loaded(self)
        return self.llama.create_completion(
            prompt,
            max_tokens=sampler.max_len_tokens,
            temperature=sampler.temp,
            top_p=sampler.top_p,
            min_p=sampler.min_p,
            frequency_penalty=sampler.frequency_penalty,
            presence_penalty=sampler.presence_penalty,
            repeat_penalty=sampler.repeat_penalty,
            top_k=sampler.top_k,
            stop=stops
        )['choices'][0]['text']
    

    def stream(
            self,
            prompt: Union[str, list[int]],
            stops: list[Union[str, int]] = [],
            sampler: SamplerSettings = DefaultSampling
        ) -> Generator:

        """
        Given a prompt, return a generator that yields dicts containing tokens.

        To get the token string itself, subscript the dict with:

        `['choices'][0]['text']`

        prompt: The text from which to generate

        The following parameters are optional:
        - stops: A list of strings and/or token IDs at which to end the generation early
        - sampler: The SamplerSettings object used to control text generation
        """

        assert isinstance(prompt, (str, list)), \
            f"stream: prompt should be string or list[int], not {type(prompt)}"
        if isinstance(prompt, list):
            assert all(isinstance(tok, int) for tok in prompt), \
                "stream: some token in prompt is not an integer"
        assert isinstance(stops, list), \
            f"stream: parameter `stops` should be a list, not {type(stops)}"
        assert all(isinstance(item, (str, int)) for item in stops), \
            f"stream: some item in parameter `stops` is not a string or int"

        if self.verbose:
            print_verbose(f'using the following sampler settings for Model.stream:')
            print_verbose(f'max_len_tokens    == {sampler.max_len_tokens}')
            print_verbose(f'temp              == {sampler.temp}')
            print_verbose(f'top_p             == {sampler.top_p}')
            print_verbose(f'min_p             == {sampler.min_p}')
            print_verbose(f'frequency_penalty == {sampler.frequency_penalty}')
            print_verbose(f'presence_penalty  == {sampler.presence_penalty}')
            print_verbose(f'repeat_penalty    == {sampler.repeat_penalty}')
            print_verbose(f'top_k             == {sampler.top_k}')
        
        # if any stop item is a token ID (int)
        if any(isinstance(stop, int) for stop in stops):
            # stop_strs is a list of all stopping strings
            stop_strs: list[str] = [stop for stop in stops if isinstance(stop, str)]
            # stop_token_ids is a list of all stop token IDs
            stop_token_ids: list[int] = [tok_id for tok_id in stops if isinstance(tok_id, int)]
            def stop_on_token_ids(tokens, *args, **kwargs):
                return tokens[-1] in stop_token_ids
            stopping_criteria = StoppingCriteriaList([stop_on_token_ids])
            assert_model_is_loaded(self)
            return self.llama.create_completion(
                prompt,
                max_tokens=sampler.max_len_tokens,
                temperature=sampler.temp,
                top_p=sampler.top_p,
                min_p=sampler.min_p,
                frequency_penalty=sampler.frequency_penalty,
                presence_penalty=sampler.presence_penalty,
                repeat_penalty=sampler.repeat_penalty,
                top_k=sampler.top_k,
                stream=True,
                stop=stop_strs,
                stopping_criteria=stopping_criteria
            )

        assert_model_is_loaded(self)
        return self.llama.create_completion(
            prompt,
            max_tokens=sampler.max_len_tokens,
            temperature=sampler.temp,
            top_p=sampler.top_p,
            min_p=sampler.min_p,
            frequency_penalty=sampler.frequency_penalty,
            presence_penalty=sampler.presence_penalty,
            repeat_penalty=sampler.repeat_penalty,
            top_k=sampler.top_k,
            stream=True,
            stop=stops
        )


    def stream_print(
            self,
            prompt: Union[str, list[int]],
            stops: list[Union[str, int]] = [],
            sampler: SamplerSettings = DefaultSampling,
            end: str = "\n",
            file: _SupportsWriteAndFlush = sys_stdout,
            flush: bool = True
    ) -> str:
        """
        Given a prompt, stream text as it is generated, and return the generated string.
        The returned string does not include the `end` parameter.

        `Model.stream_print(...)` is a shorthand for:
        ```
        s = Model.stream(prompt, stops=stops, sampler=sampler)
        for i in s:
            tok = i['choices'][0]['text']
            print(tok, end='', file=file, flush=flush)
        print(end, end='', file=file, flush=True)
        ```

        prompt: The text from which to generate

        The following parameters are optional:
        - stops: A list of strings and/or token IDs at which to end the generation early
        - sampler: The SamplerSettings object used to control text generation
        - end: A string to print after the generated text
        - file: The file where text should be printed
        - flush: Whether to flush the stream after each token
        """
        
        token_generator = self.stream(
            prompt=prompt,
            stops=stops,
            sampler=sampler
        )

        res = ''
        for i in token_generator:
            tok = i['choices'][0]['text']
            print(tok, end='', file=file, flush=flush)
            res += tok

        # always flush stream after generation is done
        print(end, end='', file=file, flush=True)

        return res


    def ingest(self, text: str) -> None:
        """
        Ingest the given text into the model's cache
        """

        assert_model_is_loaded(self)
        self.llama.create_completion(
            text,
            max_tokens=1,
            temperature=0.0
        )
    

    def next_candidates(
            self,
            prompt: str,
            k: int
        ) -> list[str]:
        """
        Given prompt `str` and k `int`, return a sorted list of the
        top k candidates for most likely next token
        """
        # WIP
        raise NotImplementedError(
            'Model.next_candidates() is not yet implemented'
        )
        assert_model_is_loaded(self)
        tokens = self.llama.tokenize(prompt.encode('utf-8', errors='ignore'))
        self.llama.eval(tokens)
        self.llama.scores

def assert_model_is_loaded(model: Model) -> None:
    """
    Ensure the Model is fully constructed, such that
    `Model.llama._model.model is not None` is guaranteed to be `True`

    Raise ModelUnloadedException otherwise
    """
    if not hasattr(model, 'llama'):
        raise ModelUnloadedException(
            "easy_llama.Model instance has no attribute 'llama'"
        )
    if not hasattr(model.llama, '_model'):
        raise ModelUnloadedException(
            "llama_cpp.Llama instance has no attribute '_model'"
        )
    if not hasattr(model.llama._model, 'model'):
        raise ModelUnloadedException(
            "llama_cpp._internals._LlamaModel instance has no attribute 'model'"
        )
    if model.llama._model.model is None:
        raise ModelUnloadedException(
            "llama_cpp._internals._LlamaModel.model is None"
        )
